"""
This module provides the starting point for the slurm node. You do not have to call
anything of this file yourself.
"""
import os

from .guard import prevent_distribution
from .function import SlurmFunction
from .function_map import set_entry_point
import json
import sys


def parse_args():
    batch_file_path = sys.argv[1]  # the file with the code (function definition)
    # determine whether function calls are provided as an argument or in a temp file.
    mode = sys.argv[2]
    if mode == "arg":
        function_calls = json.loads(sys.argv[3])
    elif mode == "temp":
        with open(sys.argv[3]) as f:
            function_calls = json.load(f)
        os.remove(sys.argv[3])  # delete the temp file
    else:
        raise ValueError("Unknown function call mode. Expected 'arg' or 'temp'.")
    assert isinstance(function_calls, list), "Expected a list of dicts"
    return batch_file_path, function_calls


def main():
    prevent_distribution()  # make sure, the code on the node does not distribute itself.
    batch_file, function_calls = parse_args()

    set_entry_point(batch_file)
    with open(batch_file, "r") as f:
        code = "".join(f.readlines())

    # Workaround as otherwise __name__ is not defined
    global __name__
    __name__ = None

    glob = dict(globals())
    glob["__file__"] = batch_file
    exec(code, glob)

    # Execute the functions
    for fc in function_calls:
        SlurmFunction.call(fc["func_id"], *fc.get("args", []), **fc.get("kwargs", {}))


if __name__ == "__main__":
    main()



"""
slurminade allows to distribute function calls to slurm using decorators.

```
import slurminade

slurminade.update_default_configuration(partition="alg")  # global options for slurm

# If no slurm environment is found, the functions are called directly to make scripts
# compatible with any environment.
# You can enforce slurm with `slurminade.set_dispatcher(slurminade.SlurmDispatcher())`

# use this decorator to make a function distributable with slurm
@slurminade.slurmify(constraint="alggen02")  # function specific options can be specified
def prepare():
    print("Prepare")

@slurminade.slurmify()
def f(foobar):
    print(f"f({foobar})")

@slurminade.slurmify()
def clean_up():
    print("Clean up")


if __name__ == "__main__":
    jid = prepare.distribute()

    with slurminade.Batch(max_size=20) as batch:  # automatically bundles up to 20 tasks
        # run 100x f after `prepare` has finished
        for i in range(100):
            f.wait_for(jid).distribute(i)

        # clean up after the previous jobs have finished
        jids = batch.flush()
        clean_up.wait_for(jids).distribute()
```

Project structure:
- batch.py: Contains code for bundling tasks, so we don't spam slurm with too many.
- conf.py: Contains code for managing the configuration of slurm.
- dispatcher.py: Contains code for actually dispatching tasks to slurm.
- execute.py: Contains code to execute the task on the slurm node.
- function.py: Contains the code for making a function slurm-compatible.
- function_map.py: Saves all the slurified functions.
- guard.py: Contains code to prevent you accidentally DDoSing your infrastructure.
- options.py: Contains a simple data structure to save slurm options.
"""

from .function import slurmify
from .conf import update_default_configuration, set_default_configuration
from .guard import set_dispatch_limit, allow_recursive_distribution
from .batch import Batch
from .dispatcher import (
    srun,
    sbatch,
    SlurmDispatcher,
    set_dispatcher,
    get_dispatcher,
    TestDispatcher,
    SubprocessDispatcher,
)
from .function_map import set_entry_point



"""
This file saves the default configuration for slurm.
"""

import json
import os.path
from pathlib import Path


def _load_default_conf():
    default_conf_file = os.path.join(Path.home(), ".slurminade_default.json")
    try:
        if os.path.isfile(default_conf_file):
            with open(default_conf_file, "r") as f:
                return json.load(f)
        else:
            return {}
    except Exception as e:
        print(
            f"slurminade could not open default configuration {default_conf_file}!\n{str(e)}"
        )
    return {}


__default_conf = _load_default_conf()


def update_default_configuration(conf=None, **kwargs):
    if conf:
        __default_conf.update(conf)
    if kwargs:
        __default_conf.update(kwargs)


def set_default_configuration(conf=None, **kwargs):
    __default_conf = {}
    update_default_configuration(conf, **kwargs)


def _get_conf(conf=None):
    conf = conf if conf else {}
    conf_ = __default_conf.copy()
    conf_.update(conf)
    return conf_



"""
Some security measures to prevent you from DDoSing your infrastructure.
1. Preventing recursive distributions, i.e., slurm nodes also distributing tasks.
2. Limiting the number of distributed tasks.

You can disable these security mechanisms by
`allow_recursive_distribution` and `set_dispatch_limit(None)`.
"""
import typing

__exec_flag = False


def guard_recursive_distribution():
    if __exec_flag:
        raise RuntimeError(
            "Cannot distribute a task within a distributed task! "
            "(maybe you forgot to guard you script with "
            "'if __name__==\"__main__\":'?)"
        )


def prevent_distribution():
    __exec_flag = True


def allow_recursive_distribution() -> None:
    """
    Allow recursive distribution. Dangerous!
    :return: None
    """
    __exec_flag = False


class TooManyDispatchesError(RuntimeError):
    def __init__(self, n_calls):
        self.n_calls = n_calls

    def __str__(self):
        return (
            f"Exceeded the dispatch limit of {self.n_calls} calls. "
            f"This limit has been introduced to prevent you from spamming your "
            f"slurm environment in case of a bug. You can increase it"
            f" using `set_dispatch_limit`."
        )


class _DispatchGuard:
    def __init__(self, max_calls):
        self.max_calls = max_calls
        self.remaining_calls = max_calls

    def __call__(self):
        if not self.max_calls:
            return None
        if self.remaining_calls <= 0:
            raise TooManyDispatchesError(self.max_calls)
        self.remaining_calls -= 1
        return self.remaining_calls

    def set_limit(self, n):
        self.max_calls = n
        self.remaining_calls = n


dispatch_guard = _DispatchGuard(100)


def set_dispatch_limit(n: typing.Optional[int]):
    """
    Set a limit to the number of dispatches. This feature has been introduced to
    prevent you from accidentally DDoSing you Slurm environment due to a bug.
    :param n: The maximal number of dispatches.
    :return: None
    """
    dispatch_guard.set_limit(n)



from setuptools import setup


def readme():
    with open("README.md") as f:
        return f.read()


setup(
    name="slurminade",
    version="0.6.0",
    description="A decorator-based slurm runner",
    long_description=readme(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    keywords="slurm",
    url="https://github.com/d-krupke/slurminade",
    author="Dominik Krupke",
    author_email="krupke@ibr.cs.tu-bs.de",
    license="MIT",
    packages=["slurminade"],
    install_requires=["simple_slurm"],
    entry_points={
        #  'console_scripts': ['slurminade-exec=slurminade.execute:main'],
    },
    python_requires=">=3.7",
    include_package_data=True,
    zip_safe=False,
)



"""
Contains code for bundling function calls together.
"""
import typing
from collections import defaultdict

from slurminade.dispatcher import Dispatcher
from slurminade.dispatcher import get_dispatcher, FunctionCall, set_dispatcher
from slurminade.function import SlurmFunction
from slurminade.options import SlurmOptions


class TaskBuffer:
    """
    A simple container to buffer all the tasks by their options.
    We can only bundle tasks with the same slurm options.
    """
    def __init__(self):
        self._tasks = defaultdict(list)

    def add(self, task: FunctionCall, options: SlurmOptions) -> int:
        self._tasks[options].append(task)
        return len(self._tasks[options])

    def items(self):
        for opt, tasks in self._tasks.items():
            if tasks:
                yield opt, tasks

    def get(self, options: SlurmOptions) -> typing.List[FunctionCall]:
        return self._tasks[options]

    def clear(self):
        self._tasks.clear()


class Batch(Dispatcher):
    """
    The logic to buffer the function calls. It wraps the original dispatcher.

    You can use
    ```
    with slurminade.Batch(max_size=20) as batch:  # automatically bundles up to 20 tasks
        # run 100x f
        for i in range(100):
            f.distribute(i)
    ```
    """
    def __init__(self, max_size: int):
        """
        :param max_size: Bundle up to this many calls.
        """
        super().__init__()
        self.max_size = max_size
        self.subdispatcher = get_dispatcher()
        self._tasks = TaskBuffer()

    def flush(
        self, options: typing.Optional[SlurmOptions] = None
    ) -> typing.Iterable[int]:
        """
        Distribute all buffered tasks. Return the job ids used.
        :param options: Only flush tasks with specific options.
        :return: A list of job ids.
        """
        job_ids = []
        if options is None:
            for opt, tasks in self._tasks.items():
                while tasks:
                    job_id = self.subdispatcher(tasks[: self.max_size], opt)
                    job_ids.append(job_id)
                    tasks = tasks[self.max_size :]

        else:
            tasks = self._tasks.get(options)
            while len(tasks) > self.max_size:
                job_id = self.subdispatcher(tasks[: self.max_size], options)
                job_ids.append(job_id)
                tasks = tasks[: self.max_size]
        self._tasks.clear()
        return job_ids

    def add(self, func: SlurmFunction, *args, **kwargs):
        """
        You can also add a task using `add` instead of `distribute`.
        :param func: The function to call
        :param args: The positional arguments
        :param kwargs: The keywords arguments.
        :return: None
        """
        self._dispatch(
            [FunctionCall(func.func_id, args, kwargs)], func.special_slurm_opts
        )

    def _dispatch(
        self, funcs: typing.Iterable[FunctionCall], options: SlurmOptions
    ) -> int:
        for func in funcs:
            self._tasks.add(func, options)
        return -1

    def srun(self, command: str, conf: dict = None, simple_slurm_kwargs: dict = None):
        return self.subdispatcher.srun(command, conf, simple_slurm_kwargs)

    def sbatch(self, command: str, conf: dict = None, simple_slurm_kwargs: dict = None):
        return self.subdispatcher.sbatch(command, conf, simple_slurm_kwargs)

    def __enter__(self):
        self.subdispatcher = get_dispatcher()
        set_dispatcher(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            print("Aborted due to exception.")
            return
        self.flush()
        set_dispatcher(self.subdispatcher)

    def __del__(self):
        self.flush()

    def is_sequential(self):
        return self.subdispatcher.is_sequential()
